
#include <string.h>
#include <stdio.h>
#include <ctype.h>

/*
Borrowed and stripped of unused bits from the Python 2.5 string library
*/

int stringview_fastsearch(const char* s, int n, const char* p, int m) {
    long mask;
    int skip, count = 0;
    int i, j, mlast, w;

    w = n - m;

    if (w < 0)
        return -1;

    /* look for special cases */
    if (m <= 1) {
        if (m <= 0)
            return -1;
        /* use special case for 1-character strings */
        for (i = 0; i < n; i++)
            if (s[i] == p[0])
                return i;
        return -1;
    }

    mlast = m - 1;

    /* create compressed boyer-moore delta 1 table */
    skip = mlast - 1;
    /* process pattern[:-1] */
    for (mask = i = 0; i < mlast; i++) {
        mask |= (1 << (p[i] & 0x1F));
        if (p[i] == p[mlast])
            skip = mlast - i - 1;
    }
    /* process pattern[-1] outside the loop */
    mask |= (1 << (p[mlast] & 0x1F));

    for (i = 0; i <= w; i++) {
        /* note: using mlast in the skip path slows things down on x86 */
        if (s[i+m-1] == p[m-1]) {
            /* candidate match */
            for (j = 0; j < mlast; j++)
                if (s[i+j] != p[j])
                    break;
            if (j == mlast) return i;

            /* miss: check if next character is part of pattern */
            if (!(mask & (1 << (s[i+m] & 0x1F))))
                i = i + m;
            else
                i = i + skip;
        } else {
            /* skip: check if next character is part of pattern */
            if (!(mask & (1 << (s[i+m] & 0x1F))))
                i = i + m;
        }
    }

    return -1;
}

int stringview_rfastsearch(const char* s, int n, const char* p, int m) {
    /* make faster! */
    int j;
    if (m == 0)
        return 0;
    
    for (j = n - m; j >= 0; --j)
        if (memcmp(s+j, p, m) == 0)
            return j;
    return -1;
}


int stringview_count(const char* s, int n, const char* p, int m) {
    long mask;
    int skip, count = 0;
    int i, j, mlast, w;

    w = n - m;

    if (w < 0)
        return -1;

    /* look for special cases */
    if (m <= 1) {
        if (m <= 0)
            return -1;
        /* use special case for 1-character strings */
        for (i = 0; i < n; i++)
            if (s[i] == p[0])
                count++;
        return count;
    }

    mlast = m - 1;

    /* create compressed boyer-moore delta 1 table */
    skip = mlast - 1;
    /* process pattern[:-1] */
    for (mask = i = 0; i < mlast; i++) {
        mask |= (1 << (p[i] & 0x1F));
        if (p[i] == p[mlast])
            skip = mlast - i - 1;
    }
    /* process pattern[-1] outside the loop */
    mask |= (1 << (p[mlast] & 0x1F));

    for (i = 0; i <= w; i++) {
        /* note: using mlast in the skip path slows things down on x86 */
        if (s[i+m-1] == p[m-1]) {
            /* candidate match */
            for (j = 0; j < mlast; j++)
                if (s[i+j] != p[j])
                    break;
            if (j == mlast) {
                /* got a match! */
                count++;
                i = i + mlast;
                continue;
            }
            /* miss: check if next character is part of pattern */
            if (!(mask & (1 << (s[i+m] & 0x1F))))
                i = i + m;
            else
                i = i + skip;
        } else {
            /* skip: check if next character is part of pattern */
            if (!(mask & (1 << (s[i+m] & 0x1F))))
                i = i + m;
        }
    }
    return count;
}

int stringview_expandtabs_size(const char* p, int size, int tabsize) {
    const char *e = p + size;
    int i=0, j=0, count=0;

    /* First pass: determine size of output string */
    i = j = 0;
    for (; p < e; p++) {
        if (*p == '\t') {
            if (tabsize > 0)
                j += tabsize - (j % tabsize);
            count++;
        } else {
            j++;
            if (*p == '\n' || *p == '\r') {
                i += j;
                j = 0;
            }
        }
    }
    if (count == 0) return -1;
    return i+j;
}

void stringview_expandtabs(const char* p, int size, char* q, int tabsize) {
    const char *e = p + size;
    int i=0, j=0;
    
    for (; p < e; p++) {
        if (*p == '\t') {
            if (tabsize > 0) {
                i = tabsize - (j % tabsize);
                j += i;
                while (i--)
                    *q++ = ' ';
            }
        } else {
            j++;
            *q++ = *p;
            if (*p == '\n' || *p == '\r')
                j = 0;
        }
    }
}

int stringview_findwhitespace(const char* s, int n, int* start, int* length) {
    int i, j;
    for (i=0;i<n;i++) {
        if (!isspace(s[i])) continue;
        j=i+1;
        while (j<n) {
            if (!isspace(s[j]))
                break;
            j++;
        }
        start[0] = i;
        length[0] = j-i;
        return 1;
    }
    return 0;
}

int stringview_rfindwhitespace(const char* s, int n, int* start, int* length) {
    int i, j;
    for (i=n-1;i>=0;i--) {
        if (!isspace(s[i])) continue;
        j=i-1;
        while (j>=0) {
            if (!isspace(s[j]))
                break;
            j--;
        }
        start[0] = j+1;
        length[0] = i-j;
        return 1;
    }
    return 0;
}

const unsigned char whitespace[] = "\t\n\x0b\x0c\r";
const int whitespace_length = 6;

/* could probably do with a faster strip operation */
int stringview_strip(const unsigned char* s, int n, const unsigned char* p, int m, int dire) {
    char strippable[256];
    int count = 0, offset = dire < 0 ? n-1 : 0;
    int i;
    
    for (i=0;i<256;i++) strippable[i] = 0;
    for (i=0;i<m;i++) strippable[p[i]] = 1;
    
    while (count < n) {
        if (!strippable[s[offset]])
            break;
        offset += dire;
        count++;
    }
    return count;
}

const int result_table[] = {1,1,0,1,0,0,0,1,1,0,0,1,0,0,0,1,1,1};

int stringview_compare(char* l, int ll, char*r, int rl, int op) {
    int cmpsize = ll<rl ? ll : rl;
    printf("%i (%i) ", __LINE__, cmpsize);
    op = op<0 ? 0 : (op>5 ? 5 : op);
    printf("%i op(%i) ", __LINE__, op);
    if (cmpsize)
        cmpsize = memcmp(l, r, cmpsize);
    printf("%i (%i) ", __LINE__, cmpsize);
    if (!cmpsize)
        cmpsize = ll<rl ? -1 : (ll>rl ? 1 : 0);
    printf("%i (%i) ", __LINE__, cmpsize);
    cmpsize = result_table[op+6*(cmpsize+1)];
    printf("%i (%i)\n", __LINE__, cmpsize);
    return cmpsize;
}

int stringview_hash(unsigned char* p, int l) {
    int len = l, x = *p << 7;
    while (--len >= 0)
        x = (1000003*x) ^ *p++;
    x ^= l;
    if (x == -1)
        x = -2;
    return x;
}

const int digit_mappings[6][256] = {
/*isalnum*/    {0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,1,1,1,1,1,1,1,1,1,1,0,0,0,0,0,0,0,1,1,1,1,1,1,1,1,1,1,1,1,1,1,1,1,1,1,1,1,1,1,1,1,1,1,0,0,0,0,0,0,1,1,1,1,1,1,1,1,1,1,1,1,1,1,1,1,1,1,1,1,1,1,1,1,1,1,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0},
/*isalpha*/    {0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,1,1,1,1,1,1,1,1,1,1,1,1,1,1,1,1,1,1,1,1,1,1,1,1,1,1,0,0,0,0,0,0,1,1,1,1,1,1,1,1,1,1,1,1,1,1,1,1,1,1,1,1,1,1,1,1,1,1,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0},
/*isdigit*/    {0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,1,1,1,1,1,1,1,1,1,1,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0},
/*islower*/    {0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,1,1,1,1,1,1,1,1,1,1,1,1,1,1,1,1,1,1,1,1,1,1,1,1,1,1,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0},
/*isspace*/    {0,0,0,0,0,0,0,0,0,1,1,1,1,1,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,1,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0},
/*isupper*/    {0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,1,1,1,1,1,1,1,1,1,1,1,1,1,1,1,1,1,1,1,1,1,1,1,1,1,1,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0}
};

/* */

int stringview_check(unsigned char* p, int l, int which) {
    int i;
    if (l == 0) return 0;
    for (i=0;i<l;i++) {
        if (digit_mappings[which][*p] == 0)
            return 0;
        p++;
    }
    return 1;
}

int stringview_istitle(const unsigned char* p, int l) {
    int cased = 0, previous_is_cased = 0;
    const unsigned char* e = p + l;
    for (; p < e; p++) {
        if (isupper(*p)) {
            if (previous_is_cased)
                return 0;
            previous_is_cased = 1;
            cased = 1;
        } else if (islower(*p)) {
            if (!previous_is_cased)
                return 0;
            previous_is_cased = 1;
            cased = 1; /* I don't believe this assignment is necessary */
        } else
            previous_is_cased = 0;
    }
    return cased;
}

void stringview_ul_translate(unsigned char* p, int l, int which, int incr) {
    const unsigned char* e = p+l;
    for (; p < e; p++) {
        if (digit_mappings[which][*p]) {
            p[0] += incr;
        }
    }
}

void stringview_swapcase(unsigned char* p, int l) {
    const unsigned char* e = p+l;
    for (; p < e; p++) {
        if (digit_mappings[3][*p]) {
            p[0] -= 32;
        } else if (digit_mappings[5][*p]) {
            p[0] += 32;
        }
    }
}

